:- ensure_loaded('heart.pl').

%Przkad tokenize_file('therm-rt.pl',Tokeny).

tokenize_file(File,Tokens) :-
	read_file_to_codes(File,Codes,[]),
	token_list(Tokens,Codes,[]).

% Przykiad uzycia: token_list(Lista,"napis",[]).
token_list([T|Rest]) --> blank0,token(T),!,token_list(Rest).
token_list([]) --> blank0.

% blank0 is a series of zero or more blanks.
%blank0 --> [B],{c_type(B,comment)},garbage,[E],!,{c_type(E,newline)}.
blank0 --> [B],{c_type(B,comment)},garbage,!,blank0.
blank0 --> [C],{c_type(C,blank)},!,blank0.
blank0 --> [C],{c_type(C,space)},!,blank0.
blank0 --> [].

garbage --> [C], {c_type(C,Result),Result \= newline},!,garbage.
garbage --> [].

% Several kinds of tokens.
% This is where lists of characters get converted into atoms or numbers.
token(T) --> str_token(S),{atom_codes(T,S)}.
token(T) --> special(L), {atom_codes(T,L)}.
token(T) --> word(W), {atom_codes(T,W)}.

% A word is a series of one or more letters.
% The rules are ordered so that we first try to gather as many
% characters into one digit_string as possible.
word([L|Rest]) --> letter(L),word(Rest).
word([L]) --> letter(L).

str_token(S) --> "'",str_garbage(G),"'",!,{flatten(G,S)}.
str_garbage([H|R]) --> word(H),!,str_garbage(R).
str_garbage([H|R]) --> special(H),!, str_garbage(R). 
str_garbage([C|R]) --> [C],str_garbage(R),{c_type(C,space)}.
str_garbage([]) --> [].


% Various kinds of characters...
special([C]) --> [C], {c_type(C,special)}.
letter(C) --> [C], {c_type(C,lowercase)}.
letter(C) --> [C], {c_type(C,numeric)}.
letter(C) --> [C], {c_type(C,uppercase)}.
letter(C) --> [C], {c_type(C,underscore)}.

c_type(Code,Y) :-
	Code =:= 10,
	Y = newline,!.
	
c_type(Code,Y) :-
	Code =:= 37,!,
	Y=comment.

c_type(Code,Y) :-
	Code =:= 39,!,
	Y=quote.	

c_type(Code,Y):-
	Code =:= 95,!,
	Y = underscore.
	
c_type(Code,Y) :-
	Code =:= 32,!,
	Y = space.
	
c_type(Code,Y):- 
	48 =< Code, 
	Code =< 57,!, 
	Y = numeric.
	
c_type(Code,Y):- 
	97 =< Code,
	Code=< 122,!,
	Y = lowercase.

c_type(Code,Y):-
	65 =< Code, 
	Code=< 90,!,
	Y = uppercase.
	
c_type(Code,Y):-
	Code =< 31,!,
	Y = blank.
	
c_type(Code,special).
	