:- ensure_loaded('DCG-tokenize.pl').

%Ex. ?- check_grammar('XTT_2/therm-rt.pl').
check_grammar(File) :-
	tokenize_file(File,Tokens),
	hmr(Tokens,[]).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%% HMR GRAMMAR %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

hmr --> required_defs(types_def),{write('Types OK'),nl},
	optional_defs(types_grp_def),{write('Type Groups OK'),nl},
	required_defs(attrs_def),{write('Attributes OK'),nl},
	optional_defs(attrs_grp_def), {write('Attribute Groups OK'),nl},
	required_defs(schms_def),{write('Schemas OK'),nl},
	required_defs(rules_def),{write('Rules OK'),nl},
	optional_defs(states_def),{write('States OK'),nl}.
	

%Here some modifications required. If base is numeric then can be numbering without ordered. 
%In any other case numbering requires domain to be ordered.
	
types_def --> [xtype],['['],
	type_entry_line(TE1),[,],
	type_entry_line(TE2),[,],
	type_entry_line(TE3),
	optional_entry_field(type_entry_line(TE4)),
	optional_entry_field(type_entry_line(TE5)),
	optional_entry_field(type_entry_line(TE6)),
	[']'],[.],!,
	{partition(var,[TE1,TE2,TE3,TE4,TE5,TE6],_,TypeBody),
	 is_set(TypeBody),
	 subset([name,base,domain],TypeBody)}.
	 
	
	
	% imitiacja zbioru: elementy xtype opisac jako "linijka(X,content), tak aby X przyierao nazwe elementy (base, domain); ilsoc linijek tyle ile elementow wymaganych, poza tym linijki opcjonalne, warunek - kazda linijka musi byc inna, czyli linijka(X,C1),linijka(Y,C2), X1\=X2
	

type_entry_line(X) --> [X],[:],str,{ X=name }.
type_entry_line(X) --> [X],[:],[symbolic],{ X=base }.
type_entry_line(X) --> [X],[:],[numeric],{ X=base }.
type_entry_line(X) --> [X],[:],lst(domain),{ X=domain }.
type_entry_line(X) --> [X],[:],[ss],{ X=scale }.
type_entry_line(X) --> [X],[:],[yes],{ X=ordered }.
type_entry_line(X) --> [X],[:],[no],{ X=ordered }.
type_entry_line(X) --> [X],[:],str,{ X=desc }.

attrs_def --> [xattr],['['],
	attr_entry_line(AE1),[,],
	attr_entry_line(AE2),[,],
	attr_entry_line(AE3),[,],
	attr_entry_line(AE4),
	optional_entry_field(attr_entry_line(AE5)),
	optional_entry_field(attr_entry_line(AE6)),
	optional_entry_field(attr_entry_line(AE7)),
	[']'],[.],!,
	{partition(var,[AE1,AE2,AE3,AE4,AE5,AE6,AE7],_,AttrBody),
	 is_set(AttrBody),
	 subset([name,class,type,comm],AttrBody)}.
	
	
attr_entry_line(X) --> [X],[:],str,{ X=name }.
attr_entry_line(X) --> [X],[:],[simple],{ X=class }.
attr_entry_line(X) --> [X],[:],[general],{ X=class }.
attr_entry_line(X) --> [X],[:],str,{ X=type }.
attr_entry_line(X) --> [X],[:],[in],{ X=comm }.
attr_entry_line(X) --> [X],[:],[out],{ X=comm }.
attr_entry_line(X) --> [X],[:],[inter],{ X=comm }.
attr_entry_line(X) --> [X],[:],[comm],{ X=comm }.
attr_entry_line(X) --> [X],[:],lst(general),{ X=callback }.
attr_entry_line(X) --> [X],[:],str,{ X=abbrev }.
attr_entry_line(X) --> [X],[:],str,{ X=desc }.




schms_def --> [xschm],
	schm_name,
	optional_field(schm_description), 
	[:], 
	schm_from,
	[=],[=],[>],
	schm_to,[.],!.
	
schm_name --> str.
schm_description --> [/],str.
schm_from --> lst(general).		
schm_to --> lst(general).			

	
rules_def --> [xrule],
	rule_id,[:],
	['['],rule_conditions,[']'],
	[=],[=],[>],
	['['],rule_decisions,[']'],
	optional_field(rule_actions),
	optional_field(rule_link),[.],!.
	
rule_actions --> [*],[*],[>],
	['['],rule_operations,[']'],!.

rule_id --> str,[/],str.

rule_conditions --> rule_condition,[,],rule_conditions.
rule_conditions --> rule_condition.
rule_condition --> str,operator,parameter.
operator --> [eq].
operator --> [neq].
operator --> [in].
operator --> [lt].
operator --> [gt].
operator --> [notin].
operator --> [subset].
operator --> [supset].
operator --> [sim].
operator --> [notsim].
parameter --> str.
parameter --> lst(general).



rule_decisions --> rule_decision.
rule_decisions --> rule_decision,[,],rule_decisions.
rule_decision --> str,[set],str.

rule_operations --> function,rule_operations.
rule_operations --> function.


rule_link --> [:],str.
rule_link --> [:],str,[/],str.

types_grp_def --> [xtpgr],['['],
	types_grp_entry(TGE1),[,],
	types_grp_entry(TGE2),
	optional_entry_field(types_grp_entry(TGE3)),
	[']'],[.],!,
	{partition(var,[TGE1,TGE2,TGE3],_,TypeGrpBody),
	 is_set(TypeGrpBody),
	 subset([name,types],TypeGrpBody)}.
	
types_grp_entry(X) --> [X],[:],str,{ X=name }.
types_grp_entry(X) --> [X],[:],str,{ X=abbrev }.
types_grp_entry(X) --> [X],[:],lst(general),{ X=types }.

attrs_grp_def --> [xatgr],['['],
	attrs_grp_entry(AGE1),[,],
	attrs_grp_entry(AGE2),
	optional_entry_field(attrs_grp_entry(AGE3)),
	[']'],[.],!,
	{partition(var,[AGE1,AGE2,AGE3],_,AttrGrpBody),
	 is_set(AttrGrpBody),
	 subset([name,types],AttrGrpBody)}.

attrs_grp_entry(X) --> [X],[:],str,{ X=name }.
attrs_grp_entry(X) --> [X],[:],str,{ X=abbrev }.
attrs_grp_entry(X) --> [X],[:],lst(general),{ X=types}.
	
states_def --> [xstat],str,[:],['['],
	str,[,],parameter,[']'],[.].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%% ADDITIONAL DEFS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

str --> [String],{atom(String)},!.
lst(Style) --> ['['],lst_args(Style),[']'],!.
lst_args(Style) --> lst_arg(Style).
lst_args(Style) --> lst_arg(Style), [,],lst_args(Style).
lst_arg(domain) -->  lst_arg(general),[/],[Weight],{number(Weight)}.
lst_arg(domain) -->  lst_arg(general).
lst_arg(general) --> range.
lst_arg(general) --> [Atom], {atom(Atom)}.
lst_arg(general) --> lst(general).

range --> [LowerLimit],[to],[UpperLimit],
	{atom_number(LowerLimit,_),atom_number(UpperLimit,_)}.
	
function --> str,['('],lst_arg(general),[')'].

required_defs(Def) --> Def, required_defs(Def),!.
required_defs(Def) --> Def.

optional_defs(Def) --> Def,!,optional_defs(Def).
optional_defs(_) --> [].

optional_field(Field) --> Field.
optional_field(_)--> [].

optional_entry_field(Field) --> [,],Field,!.
optional_entry_field(_) --> [].




