%
% $Id: examples1-rt.pl,v 1.1 2009-05-04 09:30:06 gjn Exp $
%
% Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
% HeaRT thermostat case
%
%
%     Copyright (C) 2004-6 by A. Ligeza, G. J. Nalepa
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%

:- ensure_loaded('heart.pl').

%
% Thermostat Named Types definitions
%
% FIXME:
%  - state all defaults
%  - "basic"? for basic type, any better name?
%
% HINTS
% - named types names should be plural, or general, eg. says
xtype [name: week_days,
       base: symbolic,
       %default -> scale: none,
       %default -> group: root,
       ordered: yes,		% default: no
       domain: [monday,tuesday,wednesday,thursday,friday,saturday,sunday],
%       domain: [monday,tuesday,wednesday,thursday,friday,saturday,sunday],
%      if the domain is oreder and no numbering is used we can still use it, numbering starts with 1
       domain: [monday/10,tuesday/12,wednesday/15,thursday/28,friday/4,saturday/34,sunday/3],
% domains are FULLY ordered -> unique numbers
       desc: 'some description'
       %NOPE abbrev: [mon,tue,wed,thr,fri,sat,sun]
      ]. 
xtype [name: type_of_day,
       base: symbolic,
       domain: [weekday,workday]
      ].
xtype [name: month_days,
       base: numeric,
       domain: [1 to 30,56]
       ].
xtype [name: hours,
       base: numeric,
       domain: [0 to 23]
       ].
xtype [name: biz_hours,
       base: symbolic,
       domain: [during_bizhours,not_bizhours]
      ].
xtype [name: months,
       base: symbolic,
       domain: [january,february,march,april,may,june,july,august,september,october,november,december],
       ordered: yes
      ].
xtype [name: seasons,
       base: symbolic,
       domain: [spring,summer,autumn,winter],
       ordered: yes
      ].
xtype [name: 'Temperature',
       base: numeric,
       %scale: none,
       domain: [10 to 30]
      ].

xtpgr [name:   'A General Date',
       abbrev: 'Date',
       types:  [hours,months]
      ].

%
% Thermostat Attribute definitions
%
% FIXME:
%  - state all defaults + domains
%xattr [name:  STRING, % unique name
%       class: simple|general
%       type:  STRING, % valid type name!
%       comm:  in|out|inter|comm,
%       desc:  STRING, % optional
%       abbrev: STRING % optional
%      ].

xattr [name:  day,
       class: simple,
       type:  week_days,
       comm:  in].
xattr [name:  time,
       class: general,
       type:  hours,
       comm:  in].
xattr [name:  today,
       class: simple,
       type:  type_of_day,
       comm:  inter].
xattr [name:  operation,
       class: simple,
       type:  biz_hours,
       comm:  inter
      ].
xattr [name:  month,
       class: simple,
       type:  months,
       comm:  in].
xattr [name:  season,
       class: simple,
       type:  seasons,
       comm:  inter].
xattr [name:  therm_set,    %mandatory!
       class: simple,       %mandatory!
       type: 'Temperature', %mandatory!
       comm: out            %mandatory!
       %callback: 'write_sth'
       % default = 
       % default = callback: none,
       % default = abbrev: none,
       % default = desc: none
      ].
%xattr [name: '',
%       class: ,
%       type:  ,
%       comm:  ].

% api:
%  add attribute
%  add type

xatgr [name:   'A General Date',
       abbrev: 'Date',
       types:  [date,time]
      ].

xstat(label1,att1,val).
xstat(label1,att2,val).
xstat(label1,att3,val).

xstat(label2,att1,val).
xstat(label2,att2,val).
xstat(label2,att3,val).

xstat(end,att1,val).
xstat(end,att2,val).
xstat(end,att3,val).

xstat ms: [att1,val].
xstat(dt,att2,val).
xstat(th,att3,val).

xstat start: [day,'monday'].
xstat start: [hour,13].
xstat start: [month,'january'].

% null in state!
% saving trajectory mode

xschm ms: [month] ==> [season].
xschm dt: [day] ==> [today].
xschm th: [today,hour] ==> [operation].
xschm os/'desc': [operation,season] ==> [therm_set].

% xrule syntax:
% xrule TableID/RuleID:
%  [Att1 ALSVOP1 Val1,
%   Att2 ALSVOP2 Val2,
%    . . .
%   Attn ALSVOPn Valn]
%  ==>
%  [AttX set ValX,
%    . . .
%   AttY set ExprY] % settheoretic or arithm in case on numeric atts
%  **>
%   [some_registered_action(Att1,AttN), % with legal (see scheme!) attributes
%    prolog_built_in_pred]
%  :DestTableId/DestRuleId. % optional, can also be just :DestTableId. (rule2table link)
xrule ms/1:
  [month in [january,february,december],
   month in [ala,ma,kota]]  
  ==>
  [season set summer].
xrule ms/2:
  [month in [march,april,may]]
  ==>
  [season set autumn]
  :os.
xrule ms/3:
  [month in [june,july,august]]
  ==>
  [season set winter]
  **>
  [write(month)]
  :os/3.
xrule ms/4:
  [month eq 5] %in [september,october,november]]
  ==>
  [season set winter]
  :os.

xrule dt/1:
  [day in [mon,tue,wed,thr,fri]]
  % FIXME: in case of general atts w/ ordered domain it should be possible to write
  % [mon to fri] as <mon,fri> FIXME how to support <,) and (,> ???
  ==>
  [today set workday]
  **>
  [kot,ma,ale].
xrule dt/2:
  [day in [sat,sun]]
  ==>
  [today set weekend].

xrule th/1:
  [today eq workday,
   hour gt 17]
  ==>
  [operation set not_bizhours].
xrule th/2:
  [today eq weekend,
   hour eq any]
  ==>
  [operation set not_bizhours].
xrule th/3:
  [today eq workday,
   hour lt 9]
  ==>
  [operation set not_bizhours].
xrule th/4:
  [today eq workday,
   hour  in [9 to 17]]
  ==>
  [operation set bizhours].

xrule os/1:
  [operation eq not_bizhours,
   season eq summer]
  ==>
  [therm_set set 27].
xrule os/2:
  [operation eq bizhours,
   season eq summer]
  ==>
  [therm_set set 24].
xrule os/3:
  [operation eq not_bizhours,
   season eq spring]
  ==>
  [therm_set set 15].
xrule os/4:
  [operation eq bizhours,
   season eq spring]
  ==>
  [therm_set set 20].
xrule os/5:
  [operation eq bizhours,
   season eq winter]
  ==>
  [therm_set set 18].
xrule os/6:
  [operation eq not_bizhours,
   season eq winter]
  ==>
  [therm_set set 14].
xrule os/7:
  [operation eq not_bizhours,
   season eq fall]
  ==>
  [therm_set set 16].
xrule os/8:
  [operation eq bizhours,
   season eq fall]
  ==>
  [therm_set set 20].

%FIXME three letters alsv operators

%FIXME
% - czy podajac wartosci atrubutow w warunkowej czesci reguly mozna uzywac dlugich i krotkich nazw naraz? np czy mozna zapisac regule tak:
%  * xrule dt/1:  [day in [monday,tue,wed,thr,friday]]  ==>  [today set workday].

