%
% $Id: heart-alsv.pl,v 1.30 2009-07-30 10:03:37 gjn Exp $
%
% Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
% HeaRT the ALSV(FD) handling
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- write('HeaRT: $Id: heart-alsv.pl,v 1.30 2009-07-30 10:03:37 gjn Exp $'),nl.

%%% Finite domain declaration
%%% domain(<name>,<list-of-values>).

%%% Set value definition
%%% set(<name>,<list-of-values>).

%%% Ordered Set value definition
%%% set(<name>,<list-of-values>,<order-predicate-name>).

%%% Attribute declaration
%%% attribute(<attribute-name>,<attribute-type>,<attribute-domain>).
% attribute(aDD,atomic,day_of_week).
% attribute(sDD,set,day_of_week).


%%% State definition: by listing all the values of the attributes
%%% state(<state-identifier>,<attribute>,<value>,<type>).
%%% state_time(<state-identifier>,<time>).
%%% <type> := atomic|set|...\
% state(s1,aMO,atomic,march).
% state(s2,sTM,set,[7,8,9,10,11,12,13,14,15]).


%%% Definition of facts: to be used in rule preconditions
%%% f(<attribute-type>,<attribute-name>,<relation>,<attribute-domain>)
%%% <attribute-type> := atomic|set|....
%%% For <attribute-type>=atomic <relation> := \eq, \neq, \in, \not\in
%%% For <attribute-type>=set <relation> := \eq, \neq, \subseteq,\supseteq,\sim, \not\sim

%%% To be defined: 'NULL','_',[]


%%% Satisfaction of a single fact versus state: a fact 
%%% f(<attribute-type>,<attribute-name>,<relation>,<attribute-domain>)
%%% is satisfied within some state <state> iff <relation> is valid within this state;
%%% Deatils are defined by the first rows of table1 and table2.
%%% valid(<fact>,<state>).

%%% Table1: cases of \eq, \neq, \in, \notin versus atomic StateValue in State
%valid(f(atomic,A,eq,Value),State) :-
%	state(State,A,atomic,StateValue),
%	Value == StateValue,
%	!.  
%valid(f(atomic,A,neq,Value),State) :-
%	state(State,A,atomic,StateValue),
%	Value =\= StateValue,
%	!. 
%valid(f(atomic,A,in,SetValue),State) :-
%	state(State,A,atomic,StateValue),
%	member(StateValue,SetValue),
%	!. 
%valid(f(atomic,A,notin,SetValue),State) :-
%	state(State,A,atomic,StateValue),
%	\+member(StateValue,SetValue),
%	!. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% heart

alsv_valid(Att in [L to U], State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[L]), alsv_values_check(Att,[U]),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,[L],[NormL]),
	normalize(Att,[U],[NormU]),
	NormStateValue =< NormU,
	NormStateValue >= NormL,		
	heart_debug(2,['Valid:',Att,in,[L to U]]).
	
alsv_valid(Att notin [L to U], State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[L]), alsv_values_check(Att,[U]),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,[L],[NormL]),
	normalize(Att,[U],[NormU]),
	NormStateValue >= NormU,
	NormStateValue =< NormL,	
	heart_debug(2,['Valid:',Att,notin,[L to U]]).
	
alsv_valid(Att in SetValue, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), 	
	alsv_values_check(Att,SetValue),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,SetValue,NormSetValue),
	member(NormStateValue,NormSetValue),
	heart_debug(2,['Valid:',Att,in,SetValue]).
	
alsv_valid(Att notin SetValue, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,SetValue),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,SetValue,NormSetValue),
	\+ member(NormStateValue,NormSetValue),
	heart_debug(2,['Valid:',Att,notin,SetValue]).
	
alsv_valid(Att eq any, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]),
	heart_debug(2,['Valid:',Att,eq,any]).
	
alsv_valid(Att eq any, State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	heart_debug(2,['Valid:',Att,eq,any]).

alsv_valid(Att eq null, State) :-
	%alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	\+ alsv_values_get(State,Att,StateValue), %No values for this state
	heart_debug(2,['Valid:',Att,eq,null]).
	
alsv_valid(Att eq null, State) :-
	%alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	StateValue = null, 					   %State value explicite set to null.
	heart_debug(2,['Valid:',Att,eq,null]).
	
alsv_valid(Att eq Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue == NormStateValue,
	heart_debug(2,['Valid:',Att,eq,Value]).
	
alsv_valid(Att neq Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue == NormStateValue,
	heart_debug(2,['Valid:',Att,neq,Value]).

alsv_valid(Att gt Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),	
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue < NormStateValue,
	heart_debug(2,['Valid:',Att,gt,Value]).
	
alsv_valid(Att gte Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),	
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue =< NormStateValue,
	heart_debug(2,['Valid:',Att,gte,Value]).
	
alsv_valid(Att lt Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue > NormStateValue,
	heart_debug(2,['Valid:',Att,lt,Value]).
	
alsv_valid(Att lte Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue >= NormStateValue,
	heart_debug(2,['Valid:',Att,lte,Value]).
	
%%% Table2: cases of \eq, \neq, \subseteq, \supseteq,\sim, \notsim  versus set StateValue in State
alsv_valid(Att eq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	eqset(NormSet,NormStateValue),
	heart_debug(2,['Valid:',Att,eq,Set]).
	
alsv_valid(Att neq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	neqset(NormSet,NormStateValue),
	heart_debug(2,['Valid:',Att,eq,Set]).
	
alsv_valid(Att subseteq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	subset(NormStateValue,NormSet),
	heart_debug(2,['Valid:',Att,subset,Set]).
	
alsv_valid(Att supseteq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	subset(NormSet,NormStateValue),
	heart_debug(2,['Valid:',Att,supset,Set]).
	
alsv_valid(Att sim Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	intersection(NormStateValue,NormSet,[_|_]),
	heart_debug(2,['Valid:',Att,sim,Set]).
	
alsv_valid(Att notsim Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	intersection(NormStateValue,NormSet,[]),
	heart_debug(2,['Valid:',Att,notsim,Set]).

alsv_valid(F,_) :-
	heart_debug(2,['Check for:',F,'NYI!']),
	fail.
	
alsv_values_get(State,Att,StateValue) :-
	xstat State: [Att,StateValue],!.
	
alsv_values_get(_,Att,StateValue) :-
	xattr(A),
	member(name:Att,A),
	member(comm:C,A),
	(C=in;C=inter),
	fire_callback(Att),
	xstat current: [Att,StateValue].

%valid(f(set,A,\subseteq,SetValue),State) :-
%	state(State,A,set,StateValue),
%	subset(SetValue,StateValue),
%	!.
%valid(f(set,A,\supseteq,SetValue),State) :-
%	state(State,A,set,StateValue),
%	subset(StateValue,SetValue),
%	!.
%valid(f(set,A,\sim,SetValue),State) :-
%	state(State,A,set,StateValue),
%	intersect(SetValue,StateValue,[_|_]),
%	!.
%valid(f(set,A,\notsim,SetValue),State):-
%	state(State,A,set,StateValue),
%	intersect(SetValue,StateValue,[]),
%	!.
	


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% alsvd checks
alsv_values_check(_,[]).

%set state value check
% alsv_values_check(Att,[LS to US|Rest]) :-
	% alsv_domain(Att,[LD to UD],numeric), 
	% heart_debug(2,['Check:',Att,numericreg,LS to US]),		
	% LS >= LD,
	% US =< UD,	
	% alsv_values_check(Att,Rest),
	% !.
	
%domain is set, but value X to Y
alsv_values_check(Att,[LS to US|Rest]) :-
	alsv_domain(Att,Domain,numeric), 
	heart_debug(2,['Check:',Att,numericreg,LS to US]),		
	normalize(Att,Domain,NDomain),	
	sort(NDomain, [First|SNDomain]),
	last(SNDomain,Last),
	LS >= First,
	US =< Last,
	alsv_values_check(Att,Rest),
	!.
	
% alsv_values_check(Att,[StateValue|Rest]) :-
	% nonvar(StateValue),
	% alsv_domain(Att,[LD to UD],numeric),
	% heart_debug(2,['Check:',Att,numericreg,StateValue]),
	% number(StateValue),
	% StateValue >= LD,
	% StateValue =< UD,	
	% alsv_values_check(Att,Rest),
	% !.
	
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),	
	alsv_domain(Att,Domain,numeric),
	heart_debug(2,['Check:',Att,numeric,StateValue]),	
	math_member(StateValue,Domain), % check if value in the domain with math_memeber
	alsv_values_check(Att,Rest),
	!.
	
%domain is set, but value X to Y
alsv_values_check(Att,[LS to US|Rest]) :-
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,symbolic), 
	heart_debug(2,['Check:',Att,symbolic,LS to US]),		
	normalize(Att,Domain,NDomain),	
	sort(NDomain, [First|SNDomain]),
	last(SNDomain,Last),
	LS >= First,
	US =< Last,
	alsv_values_check(Att,Rest),
	!.
		
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	number(StateValue),
	math_member(StateValue,Domain), %chceck if weight in the domain with math_memeber
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	member(StateValue/_,Domain),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	number(StateValue),
	truncate(StateValue) =:= StateValue,
	TruncatedValue is truncate(StateValue),
	nth1(TruncatedValue,Domain,_),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
%	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	member(StateValue,Domain),
	alsv_values_check(Att,Rest),
	!.

alsv_values_check(Att,[V|_]) :-
	write('  ERROR: values of \"'), write(Att),
	write('\": '), write(V), write(' NOT in the domain!'),nl,!,
	false.

%retrieves attribute domain, and base: symbolic, numeric
alsv_domain(Att,ExpandedDomain,Base) :-
	xattr(A),
	member(name:Att,A),
	member(type:T,A),
	xtype(Ts),
	member(name:T,Ts),
	member(domain:Domain,Ts),
	expand_range_expression(Domain,ExpandedDomain),
	member(base:Base,Ts).

% retrieves attribute class: simple, general
alsv_attr_class(Att,Class) :-
	xattr(A),
	member(name:Att,A),
	member(class:Class,A).

alsv_order(Att,yes) :-
	xattr(A),
	member(name:Att,A),
	member(type:_,A),
	xtype(Ts),
	member(base:numeric, Ts),!.

alsv_order(Att,Order) :-
	xattr(A),
	member(name:Att,A),
	member(type:T,A),
	xtype(Ts),
	member(name:T,Ts),
	member(ordered:Order,Ts),!.
	
normalize(Att, Value, Output) :-
	\+ alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	member(Value,Domain),
	Value = Output,!.
	
normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	number(Value),			%and change output to floats if number (can be outside the domain)
	alsv_domain(Att,Domain,_),
	math_member(Value,Domain),	
	Output is Value*1.0,!.

normalize(Att, Value, Output) :-	
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	member(Value/Numbering,Domain), 
	Output is Numbering*1.0,!.
	
normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	atom(Value),			%and change output to floats if number
	nth1(Number,Domain,Value),
	Output is Number*1.0.
	
normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	number(Value),
	truncate(Value) =:= Value,
	TruncatedNumber is truncate(Value),
	nth1(TruncatedNumber,Domain,_),	%and change output to floats
	Output is TruncatedNumber*1.0,!.
	
normalize(Att, [X to Y| Rest1], Rest2) :-
	alsv_order(Att,yes),
	normalize(Att,X,XN),
	normalize(Att,Y,YN),
	CXN is ceiling(XN),
	FYN is floor(YN),
	findall(FV,(between(CXN,FYN,V),FV is V*1.0),InsideList),		
	merge([XN|InsideList],[YN],DuplicatedList),
	list_to_set(DuplicatedList,List),
	alsv_domain(Att,Domain,_),
	normalize(Att,Domain,NDomain),						% Due to expand_range_exp we are sure tht no inf loop occure
	intersection(NDomain, List, FilteredList),
	normalize(Att, Rest1, Rest),
	merge(FilteredList, Rest,Rest2),
	!.

normalize(Att,[Number|Rest1],[Value|Rest2]) :-
	normalize(Att,Number,Value),
	normalize(Att,Rest1,Rest2),
	!.
	
normalize(_,[],[]).

%only for mixed domains ex [12,3,7 to 10, 44].
expand_range_expression([],[]).
expand_range_expression([X to Y|Values],Expanded) :-
	Start is ceiling(X),
	Stop is floor(Y),
	findall(V,(between(Start,Stop,V)),InsideList),	
	merge([X|InsideList],[Y],DuplicatedList),
	list_to_set(DuplicatedList,List),
	expand_range_expression(Values,Rest),
	merge(List,Rest,Expanded),!.
expand_range_expression([H|Values],[H|Expanded]) :-
	nonvar(H),
	expand_range_expression(Values,Expanded),!.
	

	
	
%normalize(Att, Value, Output) :-
%	alsv_order(Att,yes),
%	findall(Numbering,member(Name/Numbering,Value),Output), %and change output to floats
%	Output \= [],
%	!.
	

math_member(Weight,[_/DomainWeight|[]]) :-
	number(Weight),
	number(DomainWeight),
	Weight =:= DomainWeight. 
math_member(Weight,[_/DomainWeight|Rest]) :-
	number(Weight),
	number(DomainWeight),
	Weight =\= DomainWeight, 
	math_member(Weight,Rest).
	
math_member(Weight,[_/DomainWeight|_]) :-
	number(Weight),
	number(DomainWeight),
	Weight =:= DomainWeight,!.
	
math_member(Number,[Head|[]]) :-
	number(Number),
	number(Head),
	Number =:= Head. 
math_member(Number,[Head|Rest]) :-
	number(Number),
	number(Head),
	Number =\= Head, 
	math_member(Number,Rest).
	
math_member(Number,[Head|_]) :-
	number(Number),
	number(Head),
	Number =:= Head,!.

	


%%% Auxiliary predictaes
%%% Set equality definition:
eqset(Set1,Set2) :-
	subset(Set1,Set2),
	subset(Set2,Set1).
%%% Set in-equality definition:
neqset(Set1,Set2) :-
	\+ eqset(Set1,Set2).
	
% heart
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



% heart
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

alsv_transition(Att set Val) :-
%	atom(Val),
	\+ Val = [_|_],
	alsv_attr_class(Att,simple),
	heart_debug(2,['Adding new fact: ',Att,' set to ',Val]),
	alsv_new_val(Att,Val).

alsv_transition(Att set Val) :-
%	atom(Val),
	Val = [_|_],
	alsv_attr_class(Att,general),
%	alsv_values_check(Att,[Val]),
	alsv_new_val(Att,Val).

% fixme: no direct assert/retract, but set/get_att_val
alsv_new_val(A,V):-
%	xstat current:[A,O],
	retractall(xstat current:[A,_]),
	assert(xstat current:[A,V]),
	heart_debug(2,['Trans:',A,to,V]).
%	heart_debug(['Transition: ',A,from,O,to,V]).

%alsv_transition(Att set Val) :-
%	Val = [_|_],
	
%  [therm_set set 24].
%  [therm_set set [24,5]].

%  [therm_set set seval union([24,5],intersect('att_ala',[4 to 8])). EXCEPT COMPLEMENT? % binary relations! result=set

%  [therm_set set aeval 4 + 7 * 'att_ala']. %dowolone wyr oblicznae przez is, maths builtin, defined funcs? % result=number binary relations! therm_set in numeric

  
%%% Checking if a list of facts -- precondition formula is satisfied 
%%% satisfied(<list-of-facts>,State).

%satisfied([],_) :- !.
%satisfied([Fact|List],State) :-
%	valid(Fact,State),
%	satisfied(List,State).

%%% Auxiliary predictaes
%%% Set equality definition:
%eqset(Set1,Set2) :-
%	subset(Set1,Set2),
%	subset(Set2,Set1).
%%% Set in-equality definition:
%neqset(Set1,Set2) :-
%	\+ eqset(Set1,Set2).




