assertState([],[]) :- !.
assertState([Attr| Attrs],[Value|Values]) :- 
  assertz(xstat vstate:[Attr,Value]),  assertState(Attrs,Values).

retractState([],[]) :- !.
retractState([Attr| Attrs],[Value|Values]) :- 
  retract(xstat vstate:[Attr,Value]),  retractState(Attrs,Values).

retractState :- retractall(xstat vstate:[_,_]).

writeState([],[],[]) :- !.
writeState([C|Cs],[V|Vs],[[C,ExpandedV]|List]) :- 
  expand(C,V,ExpandedV),
  writeState(Cs,Vs,List).

expand(_,[],[]) :- !.  
expand(C,[H1|V],[H2|EV]) :- 
	vnormalize(C,H1,H2),!,
	expand(C,V,EV).
expand(C,V,EV) :-
	vnormalize(C,V,EV),!.


append(L,[],[L]):-!.
append([],L,L):-!.
append(Elem, List, [Elem|List]):-!.

appendList([], L, L) :- !.
appendList([Elem|L1], L2, [Elem|L3]) :- appendList(L1, L2, L3).

appendDistinct([], L, L) :- !.
appendDistinct([Elem|L1], L2, [Elem|L3]) :- 
  \+member(Elem,L2), appendDistinct(L1, L2, L3),!.
appendDistinct([Elem|L1], L2, L3) :- 
  member(Elem,L2), appendDistinct(L1, L2, L3).

lastElem(Elem, List) :- appendList(_,[Elem],List).

next(X, Y) :- number(Y), X is Y - 1,!. 
next(X, Y) :- number(X), Y is X + 1,!.

vcondition(Att eq any, Att in Domain) :- 
  alsv_domain(Att,Domain,_), !.
vcondition(Att lt Value, Att in [LB to UB]) :- 
  alsv_domain(Att,Domain,numeric),
  vnormalize(Att, Domain, [LB|_]),
  vnormalize(Att, [Value], [V]),
  next(UB, V), !.
vcondition(Att gt Value, Att in [LB to UB]) :- 
  alsv_domain(Att,Domain,numeric),
  vnormalize(Att, Domain, NDomain),
  lastElem(UB,NDomain),
  vnormalize(Att, [Value], [V]),  
  next(V,LB), !.
vcondition(C,C).
 
vnormalize(_,[LB to LB], [FLB]) :- FLB is LB*1.0,!.
vnormalize(_,[LB to UB], Output) :- 
	next(LB, P), 
	P =< UB, 
	vnormalize(_,[P to UB], Output2), 
	FLB is LB*1.0,
	append(FLB, Output2, Output),!.  
vnormalize(Att, Value, Output) :- normalize(Att, Value, Output), !.

vcomplement(Attr, V, VC):-
  vnormalize(Attr, V, NV),
  alsv_domain(Attr, Domain, _),
  vnormalize(Attr, Domain, NDomain),
  subtract(NDomain,NV, VC).

split_decision(D**>A:T/R,[D,A,T,R]) :- !.
split_decision(D**>A:T,[D,A,T,vnone]) :- !.
split_decision(D**>A,[D,A,vnone,vnone]) :- !.
split_decision(D:T/R,[D,[],T,R]) :- !.
split_decision(D:T,[D,[],T,vnone]) :- !.
split_decision(D,[D,[],vnone,vnone]) :- !.

vreport(N,D) :-
	flag_vreport(L),
	vid(ID),
	assert(xhalv(ID: D)),
	N =< L,
	vreport_process(D),!.
vreport(_,_).

vreport_process([]) :- nl,!.
vreport_process([H|T]):-
	write(H), write(' '),!,
	vreport_process(T).
	
get_current_vreport(VReport) :-
	vid(ID),
	findall(R,xhalv(ID:R),VReport),!.

get_current_vreport([]).

	
vgenid(ID) :-
	vid(ID),
	\+xhalv(ID : _),!.
	
vgenid(_) :-
	repeat,
	gensym(v,IDCandidate),
	retractall(vid(_)),
	assert(vid(IDCandidate)),
	\+xhalv(IDCandidate: _),!.

vconcat([String],String) :- !.
vconcat([E1,E2|List], String) :- concat(E1,E2,Elem), vconcat([Elem|List], String).

vconcat_ws(_, [String], String) :- !.
vconcat_ws(Separator, [E1,E2|List], String):- vconcat([E1,Separator,E2], Elem),vconcat_ws(Separator,[Elem|List], String).